from numpy import ndarray

from pyfem.elements.BaseElement import BaseElement
from pyfem.elements.IsoElementShape import IsoElementShape
from pyfem.elements.SolidPlaneSmallStrain import SolidPlaneSmallStrain
from pyfem.io.Dof import Dof
from pyfem.io.Material import Material
from pyfem.io.Section import Section
from pyfem.materials.BaseMaterial import BaseMaterial
from pyfem.utils.colors import error_style

element_data_dict = {
    'SolidPlaneStressSmallStrain': SolidPlaneSmallStrain,
    'SolidPlaneStrainSmallStrain': SolidPlaneSmallStrain
}


def get_element_data(element_id: int,
                     iso_element_shape: IsoElementShape,
                     connectivity: ndarray,
                     node_coords: ndarray,
                     section: Section,
                     dof: Dof,
                     material: Material,
                     material_data: BaseMaterial) -> BaseElement:
    class_name = f'{section.category}{section.type}{section.option}'.strip().replace(' ', '')

    if class_name in element_data_dict:
        return element_data_dict[class_name](element_id=element_id,
                                             iso_element_shape=iso_element_shape,
                                             connectivity=connectivity,
                                             node_coords=node_coords,
                                             section=section,
                                             dof=dof,
                                             material=material,
                                             material_data=material_data)
    else:
        error_msg = f'{class_name} element is not supported.\n'
        error_msg += f'The allowed element types are {list(element_data_dict.keys())}.'
        raise NotImplementedError(error_style(error_msg))


if __name__ == "__main__":
    pass



import sys
from argparse import ArgumentParser, Namespace, SUPPRESS


def get_arguments() -> Namespace:
    logo = r"""
                     ____             
        ____  __  __/ __/__  ____ ___ 
       / __ \/ / / / /_/ _ \/ __ `__ \
      / /_/ / /_/ / __/  __/ / / / / /
     / .___/\__, /_/  \___/_/ /_/ /_/ 
    /_/    /____/                     
    """

    print(logo)

    # 创建一个 argparse 解析器对象
    parser = ArgumentParser(add_help=False)

    # 添加程序输入文件选项
    parser.add_argument('-i', metavar='input', type=str,
                        help='Identify the input file.')

    # 添加程序输出文件选项
    parser.add_argument('-o', metavar='output', type=str,
                        help='Identify the output file.')

    # 添加参数选项
    parser.add_argument('-p', metavar='parameter', type=str,
                        help='Parameter to pass to the program.')

    # 添加帮助选项
    parser.add_argument('-h', '--help', action='help', default=SUPPRESS,
                        help='Show this help message and exit.')

    # 添加版本选项
    parser.add_argument('-v', '--version', action='version', help='Show program\'s version number and exit.',
                        version='pyfem 0.0.8')

    # 解析命令行参数
    args = parser.parse_args()

    # 如果未指定程序输入文件，则打印帮助并退出
    if not args.i:
        print('--------------------------------------')
        print('>>> error: the input file is required.')
        print('--------------------------------------')
        parser.print_help()
        sys.exit()

    return args



from xml.etree.ElementTree import ElementTree, Element, SubElement

from pyfem.assembly.Assembly import Assembly
from pyfem.io.Properties import Properties


def write_vtk(props: Properties, assembly: Assembly):
    root = Element("VTKFile", {
        "type": "UnstructuredGrid",
        "version": "0.1",
        "byte_order": "LittleEndian"
    })

    ugrid = SubElement(root, "UnstructuredGrid")
    piece = SubElement(ugrid, "Piece", {
        "NumberOfPoints": str(len(props.nodes)),
        "NumberOfCells": str(len(props.elements))
    })

    # 添加节点数据
    point_data = SubElement(piece, "PointData")
    temp = SubElement(point_data, "DataArray", {
        "type": "Float64",
        "Name": "Temperature",
        "NumberOfComponents": "1",
        "format": "ascii"
    })
    temp.text = ""
    for _ in props.nodes.items():
        temp.text += "0.0\n"  # 在这里添加具体的场量数值

    disp = SubElement(point_data, "DataArray", {
        "type": "Float64",
        "Name": "Displacement",
        "NumberOfComponents": "3",
        "format": "ascii"
    })
    disp.text = ""
    for u1, u2 in assembly.dof_solution.reshape(-1, 2):
        disp.text += f"{u1} {u2} 0.0 \n"

    for field_name, field_values in assembly.field_variables.items():
        field = SubElement(point_data, "DataArray", {
            "type": "Float64",
            "Name": field_name,
            "NumberOfComponents": "1",
            "format": "ascii"
        })
        field.text = ""
        for value in field_values:
            field.text += f"{value}\n"

    #
    points = SubElement(piece, "Points")
    node_coords = SubElement(points, "DataArray", {
        "type": "Float64",
        "NumberOfComponents": "3",
        "format": "ascii"
    })
    node_coords.text = ""
    for _, coord in props.nodes.items():
        if len(coord) == 2:
            node_coords.text += " ".join("{:.6f}".format(c) for c in coord) + " 0.0\n"
        else:
            node_coords.text += " ".join("{:.6f}".format(c) for c in coord) + "\n"

    #
    cells = SubElement(piece, "Cells")
    conn_elem = SubElement(cells, "DataArray", {
        "type": "Int32",
        "Name": "connectivity",
        "format": "ascii"
    })
    offset_elem = SubElement(cells, "DataArray", {
        "type": "Int32",
        "Name": "offsets",
        "format": "ascii"
    })
    types_elem = SubElement(cells, "DataArray", {
        "type": "UInt8",
        "Name": "types",
        "format": "ascii"
    })
    conn_elem.text = ""
    offset_elem.text = ""
    types_elem.text = ""
    offset = 0
    for _, connectivity in props.elements.items():
        conn_elem.text += " ".join(str(node_id) for node_id in connectivity) + "\n"
        offset += len(connectivity)
        offset_elem.text += "{}\n".format(offset)
        types_elem.text += "9\n"  # 12表示六面体单元类型

    tree = ElementTree(root)

    job_name = props.input_file.stem

    output_file = props.work_path.joinpath(job_name + '.vtu')

    tree.write(output_file, xml_declaration=True, encoding='utf-8')


if __name__ == "__main__":
    pass



from pyfem.bc.BaseBC import BaseBC
from pyfem.bc.DirichletBC import DirichletBC
from pyfem.io.BC import BC
from pyfem.io.Dof import Dof
from pyfem.mesh.NodeSet import NodeSet
from pyfem.utils.colors import error_style

bc_data_dict = {
    'DirichletBC': DirichletBC
}


def get_bc_data(bc: BC,
                dof: Dof,
                nodes: NodeSet) -> BaseBC:
    class_name = f'{bc.type}'.strip().replace(' ', '')

    if class_name in bc_data_dict:
        return bc_data_dict[class_name](bc=bc,
                                        dof=dof,
                                        nodes=nodes)
    else:
        error_msg = f'{class_name} bc is not supported.\n'
        error_msg += f'The allowed bc types are {list(bc_data_dict.keys())}.'
        raise NotImplementedError(error_style(error_msg))


if __name__ == "__main__":
    pass



from numpy import array

from pyfem.bc.BaseBC import BaseBC
from pyfem.io.BC import BC
from pyfem.io.Dof import Dof
from pyfem.mesh.NodeSet import NodeSet
from pyfem.utils.colors import error_style
from pyfem.utils.visualization import object_dict_to_string_ndarray


class DirichletBC(BaseBC):
    def __init__(self, bc: BC, dof: Dof, nodes: NodeSet) -> None:
        super().__init__(bc, dof, nodes)
        self.bc: BC = bc
        self.dof: Dof = dof
        self.nodes: NodeSet = nodes
        self.create_dof_values()

    def to_string(self, level: int = 1) -> str:
        return object_dict_to_string_ndarray(self, level)

    def show(self) -> None:
        print(self.to_string())

    def create_dof_values(self) -> None:
        bc_node_sets = self.bc.node_sets
        bc_node_ids = []
        for bc_node_set in bc_node_sets:
            bc_node_ids += self.nodes.node_sets[bc_node_set]

        # 如果发现施加当前边界条件的点集中有重复的点则抛出异常
        if len(bc_node_ids) != len(set(bc_node_ids)):
            error_msg = f'{type(self).__name__} {self.bc.name} contains repeat nodes\n'
            error_msg += f'Please check the input file'
            raise NotImplementedError(error_style(error_msg))
        else:
            self.bc_node_ids = array(bc_node_ids)

        # 确定施加的边界条件对应的全局自由度编号
        bc_dof_names = self.bc.dof
        dof_names = self.dof.names
        dof_ids = []
        for node_index in self.nodes.get_indices_by_ids(list(self.bc_node_ids)):
            for _, bc_dof_name in enumerate(bc_dof_names):
                dof_ids.append(node_index * len(dof_names) + dof_names.index(bc_dof_name))
        self.dof_ids = array(dof_ids)

        bc_value = self.bc.value
        if isinstance(bc_value, float):
            self.dof_values = array([bc_value for _ in self.dof_ids])


if __name__ == "__main__":
    from pyfem.io.Properties import Properties

    props = Properties()
    props.read_file(r'F:\Github\pyfem\examples\rectangle\rectangle.toml')
    props.verify()

    bc_data = DirichletBC(props.bcs[1], props.dof, props.nodes)
    bc_data.create_dof_values()
    bc_data.show()



import os

IS_COLORED = True

if os.environ.get('TERM', '') != '' or IS_COLORED:
    CYAN = '\033[36m'
    MAGENTA = '\033[35m'
    BLUE = '\033[34m'
    GREEN = '\033[32m'
    YELLOW = '\033[33m'
    RED = '\033[31m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    END = '\033[0m'
else:
    CYAN = ''
    MAGENTA = ''
    BLUE = ''
    GREEN = ''
    YELLOW = ''
    RED = ''
    BOLD = ''
    UNDERLINE = ''
    END = ''


def error_style(error_msg: str) -> str:
    return RED + BOLD + error_msg + END


def info_style(info_msg: str) -> str:
    return GREEN + BOLD + info_msg + END


def warn_style(info_msg: str) -> str:
    return YELLOW + BOLD + info_msg + END



# -*- coding: utf-8 -*-
"""

"""
from pyfem.Job import Job
from pyfem.io.arguments import get_arguments
from pyfem.utils.wrappers import show_running_time


@show_running_time
def main() -> None:
    args = get_arguments()

    job = Job(args.i)

    job.run()



from numpy import ndarray

from pyfem.utils.colors import GREEN, BLUE, END


def insert_spaces(n: int, text: str) -> str:
    lines = text.split('\n')
    indented_lines = [' ' * n + line for line in lines]
    return '\n'.join(indented_lines)


def object_dict_to_string(obj, level: int = 1) -> str:
    msg = BLUE + obj.__str__() + END
    msg += '\n'
    for key, item in obj.__dict__.items():
        msg += '  ' * level + f'|- {key}: {item}\n'
    return msg[:-1]


def object_dict_to_string_ndarray(obj, level: int = 1) -> str:
    msg = BLUE + obj.__str__() + END
    msg += '\n'
    for key, item in obj.__dict__.items():
        if isinstance(item, ndarray):
            msg += '  ' * level + GREEN + f'|- {key}: ' + END + f'{type(item)} with shape = {item.shape} \n'
            msg += insert_spaces(5 + (level - 1) * 2, f'{item}') + '\n'
        else:
            msg += '  ' * level + GREEN + f'|- {key}: ' + END + f'{item}\n'
    return msg[:-1]


def object_dict_to_string_assembly(obj, level: int = 1) -> str:
    msg = BLUE + obj.__str__() + END
    msg += '\n'
    for key, item in obj.__dict__.items():
        if isinstance(item, list) and len(item) > 8:
            msg += '  ' * level + GREEN + f'|- {key}: ' + END + f'{type(item)} of with length = {len(item)} \n'
        elif isinstance(item, ndarray):
            msg += '  ' * level + GREEN + f'|- {key}: ' + END + f'{type(item)} with shape = {item.shape} \n'
        elif key == 'global_stiffness':
            msg += '  ' * level + GREEN + f'|- {key}: ' + END + f'{type(item)} with shape = {item.shape} \n'
        else:
            msg += '  ' * level + GREEN + f'|- {key}: ' + END + f'{item}\n'
    return msg[:-1]



from typing import Optional

from numpy import ndarray, dot, empty

from pyfem.io.Material import Material
from pyfem.utils.visualization import object_dict_to_string_ndarray


class BaseMaterial:
    def __init__(self, material: Material, dimension: int, option: Optional[str] = None) -> None:
        self.material: Material = material
        self.dimension: int = dimension
        self.option: Optional[str] = option
        self.ddsdde: ndarray = empty(0)

    def to_string(self, level: int = 1) -> str:
        return object_dict_to_string_ndarray(self, level)

    def show(self) -> None:
        print(self.to_string())

    def get_stress(self, strain: ndarray) -> ndarray:
        sigma = dot(self.ddsdde, strain)
        return sigma

    def get_tangent(self) -> ndarray:
        return self.ddsdde



from typing import Tuple, Callable

from numpy import (empty, meshgrid, outer, column_stack, array, ndarray)
from numpy.polynomial.legendre import leggauss

from pyfem.elements.IsoElementDiagram import IsoElementDiagram
from pyfem.utils.colors import error_style
from pyfem.utils.visualization import object_dict_to_string_ndarray


class IsoElementShape:
    """
    等参单元类，设置等参单元的形函数和积分点等信息
    """
    allowed_element_type = ['empty', 'line2', 'line3', 'tria3', 'quad4', 'quad8', 'tetra4', 'hex8']

    def __init__(self, element_type: str) -> None:
        """
        当前支持的单元类型 ['empty', 'line2', 'line3', 'tria3', 'quad4', 'quad8', 'tetra4', 'hex8']
        """
        self.element_type: str = ''
        self.diagram: str = ''
        self.dimension: int = 0
        self.nodes_number: int = 0
        self.order: int = 0
        self.shape_function: Callable = get_shape_empty
        self.gp_number = 0
        self.gp_coords: ndarray = empty(0)
        self.gp_weights: ndarray = empty(0)
        self.gp_shape_values: ndarray = empty(0)
        self.gp_shape_gradients: ndarray = empty(0)

        if element_type == 'empty':
            self.element_type = element_type
        elif element_type == 'line2':
            self.element_type = element_type
            self.set_line2()
        elif element_type == 'line3':
            self.element_type = element_type
            self.set_line3()
        elif element_type == 'tria3':
            self.element_type = element_type
            self.set_tria3()
        elif element_type == 'quad4':
            self.element_type = element_type
            self.set_quad4()
        elif element_type == 'quad8':
            self.element_type = element_type
            self.set_quad8()
        elif element_type == 'tetra4':
            self.element_type = element_type
            self.set_tetra4()
        elif element_type == 'hex8':
            self.element_type = element_type
            self.set_hex8()
        else:
            error_msg = f'Unsupported element type {element_type}'
            raise NotImplementedError(error_style(error_msg))

    def to_string(self, level: int = 1) -> str:
        return object_dict_to_string_ndarray(self, level)

    def show(self) -> None:
        print(self.to_string())

    def set_line2(self) -> None:
        self.dimension = 1
        self.nodes_number = 2
        self.order = 1
        self.gp_coords, self.gp_weights = get_gauss_points(dimension=self.dimension, order=self.order)
        self.gp_number = len(self.gp_weights)
        self.shape_function = get_shape_line2
        gp_shape_values = []
        gp_shape_gradients = []
        for gp_coord in self.gp_coords:
            h, dhdxi = self.shape_function(gp_coord)
            gp_shape_values.append(h)
            gp_shape_gradients.append(dhdxi)
        self.gp_shape_values = array(gp_shape_values)
        self.gp_shape_gradients = array(gp_shape_gradients)
        self.diagram = IsoElementDiagram.line2

    def set_line3(self) -> None:
        self.dimension = 1
        self.nodes_number = 3
        self.order = 2
        self.gp_coords, self.gp_weights = get_gauss_points(dimension=self.dimension, order=self.order)
        self.gp_number = len(self.gp_weights)
        self.shape_function = get_shape_line3
        gp_shape_values = []
        gp_shape_gradients = []
        for gp_coord in self.gp_coords:
            h, dhdxi = self.shape_function(gp_coord)
            gp_shape_values.append(h)
            gp_shape_gradients.append(dhdxi)
        self.gp_shape_values = array(gp_shape_values)
        self.gp_shape_gradients = array(gp_shape_gradients)
        self.diagram = IsoElementDiagram.line3

    def set_quad4(self) -> None:
        self.dimension = 2
        self.nodes_number = 4
        self.order = 2
        self.gp_coords, self.gp_weights = get_gauss_points(dimension=self.dimension, order=self.order)
        self.gp_number = len(self.gp_weights)
        self.shape_function = get_shape_quad4
        gp_shape_values = []
        gp_shape_gradients = []
        for gp_coord in self.gp_coords:
            h, dhdxi = self.shape_function(gp_coord)
            gp_shape_values.append(h)
            gp_shape_gradients.append(dhdxi)
        self.gp_shape_values = array(gp_shape_values)
        self.gp_shape_gradients = array(gp_shape_gradients)
        self.diagram = IsoElementDiagram.quad4

    def set_quad8(self) -> None:
        self.dimension = 2
        self.nodes_number = 8
        self.order = 3
        self.gp_coords, self.gp_weights = get_gauss_points(dimension=self.dimension, order=self.order)
        self.gp_number = len(self.gp_weights)
        self.shape_function = get_shape_quad8
        gp_shape_values = []
        gp_shape_gradients = []
        for gp_coord in self.gp_coords:
            h, dhdxi = self.shape_function(gp_coord)
            gp_shape_values.append(h)
            gp_shape_gradients.append(dhdxi)
        self.gp_shape_values = array(gp_shape_values)
        self.gp_shape_gradients = array(gp_shape_gradients)
        self.diagram = IsoElementDiagram.quad8

    def set_tria3(self) -> None:
        self.dimension = 2
        self.nodes_number = 3
        self.order = 1
        self.gp_coords, self.gp_weights = get_gauss_points_triangle(order=self.order)
        self.gp_number = len(self.gp_weights)
        self.shape_function = get_shape_tria3
        gp_shape_values = []
        gp_shape_gradients = []
        for gp_coord in self.gp_coords:
            h, dhdxi = self.shape_function(gp_coord)
            gp_shape_values.append(h)
            gp_shape_gradients.append(dhdxi)
        self.gp_shape_values = array(gp_shape_values)
        self.gp_shape_gradients = array(gp_shape_gradients)
        self.diagram = IsoElementDiagram.tria3

    def set_tetra4(self) -> None:
        self.dimension = 3
        self.nodes_number = 4
        self.order = 1
        self.gp_coords, self.gp_weights = get_gauss_points_tetra(order=self.order)
        self.gp_number = len(self.gp_weights)
        self.shape_function = get_shape_tetra4
        gp_shape_values = []
        gp_shape_gradients = []
        for gp_coord in self.gp_coords:
            h, dhdxi = self.shape_function(gp_coord)
            gp_shape_values.append(h)
            gp_shape_gradients.append(dhdxi)
        self.gp_shape_values = array(gp_shape_values)
        self.gp_shape_gradients = array(gp_shape_gradients)
        self.diagram = IsoElementDiagram.tetra4

    def set_hex8(self) -> None:
        self.dimension = 3
        self.nodes_number = 8
        self.order = 2
        self.gp_coords, self.gp_weights = get_gauss_points(dimension=self.dimension, order=self.order)
        self.gp_number = len(self.gp_weights)
        self.shape_function = get_shape_hex8
        gp_shape_values = []
        gp_shape_gradients = []
        for gp_coord in self.gp_coords:
            h, dhdxi = self.shape_function(gp_coord)
            gp_shape_values.append(h)
            gp_shape_gradients.append(dhdxi)
        self.gp_shape_values = array(gp_shape_values)
        self.gp_shape_gradients = array(gp_shape_gradients)
        self.diagram = IsoElementDiagram.hex8


def get_shape_empty(xi: ndarray) -> Tuple[ndarray, ndarray]:
    h = empty(0)
    dhdxi = empty(shape=(0, 0))

    return h, dhdxi


def get_shape_line2(xi: ndarray) -> Tuple[ndarray, ndarray]:
    """
    两节点直线单元
    """

    # 0---------------1
    #         +-->x0

    if len(xi) != 1:
        raise NotImplementedError(error_style(f'coordinate {xi} must be dimension 1'))

    h = empty(2)
    dhdxi = empty(shape=(2, 1))

    h[0] = 0.5 * (1.0 - xi)
    h[1] = 0.5 * (1.0 + xi)

    dhdxi[0, 0] = -0.5
    dhdxi[1, 0] = 0.5

    return h, dhdxi


def get_shape_line3(xi: ndarray) -> Tuple[ndarray, ndarray]:
    """
    三节点直线单元
    """

    # 0-------1-------2
    #         +-->x0

    if len(xi) != 1:
        raise NotImplementedError(error_style(f'coordinate {xi} must be dimension 1'))

    h = empty(3)
    dhdxi = empty(shape=(1, 3))

    h[0] = 0.5 * (1.0 - xi) - 0.5 * (1.0 - xi * xi)
    h[1] = 1 - xi * xi
    h[2] = 0.5 * (1.0 + xi) - 0.5 * (1.0 - xi * xi)

    dhdxi[0, 0] = -0.5 + xi
    dhdxi[0, 1] = -2.0 * xi
    dhdxi[0, 2] = 0.5 + xi

    return h, dhdxi


def get_shape_tria3(xi: ndarray) -> Tuple[ndarray, ndarray]:
    """
    三节点三角形单元
    """

    # 2
    # * *
    # *   *
    # *     *
    # x1      *
    # |         *
    # 0--x0 * * * 1

    if len(xi) != 2:
        raise NotImplementedError(error_style(f'coordinate {xi} must be dimension 2'))

    h = empty(3)
    dhdxi = empty(shape=(3, 2))
    xi = xi

    h[0] = 1.0 - xi[0] - xi[1]
    h[1] = xi[0]
    h[2] = xi[1]

    dhdxi[0, 0] = -1.0
    dhdxi[1, 0] = 1.0
    dhdxi[2, 0] = 0.0

    dhdxi[0, 1] = -1.0
    dhdxi[1, 1] = 0.0
    dhdxi[2, 1] = 1.0

    return h, dhdxi


def get_shape_quad4(xi: ndarray) -> Tuple[ndarray, ndarray]:
    """
    四节点四边形单元
    """

    # 3---------------2
    # |       x1      |
    # |       |       |
    # |       o--x0   |
    # |               |
    # |               |
    # 0---------------1

    if len(xi) != 2:
        raise NotImplementedError(error_style(f'coordinate {xi} must be dimension 2'))

    h = empty(4)
    dhdxi = empty(shape=(4, 2))

    h[0] = 0.25 * (1.0 - xi[0]) * (1.0 - xi[1])
    h[1] = 0.25 * (1.0 + xi[0]) * (1.0 - xi[1])
    h[2] = 0.25 * (1.0 + xi[0]) * (1.0 + xi[1])
    h[3] = 0.25 * (1.0 - xi[0]) * (1.0 + xi[1])

    dhdxi[0, 0] = -0.25 * (1.0 - xi[1])
    dhdxi[1, 0] = 0.25 * (1.0 - xi[1])
    dhdxi[2, 0] = 0.25 * (1.0 + xi[1])
    dhdxi[3, 0] = -0.25 * (1.0 + xi[1])

    dhdxi[0, 1] = -0.25 * (1.0 - xi[0])
    dhdxi[1, 1] = -0.25 * (1.0 + xi[0])
    dhdxi[2, 1] = 0.25 * (1.0 + xi[0])
    dhdxi[3, 1] = 0.25 * (1.0 - xi[0])

    return h, dhdxi


def get_shape_quad8(xi: ndarray) -> Tuple[ndarray, ndarray]:
    """
    八节点四边形单元
    """

    # 3-------6-------2
    # |       x1      |
    # |       |       |
    # 7       o--x0   5
    # |               |
    # |               |
    # 0-------4-------1

    if len(xi) != 2:
        raise NotImplementedError(error_style(f'coordinate {xi} must be dimension 2'))

    h = empty(8)
    dhdxi = empty(shape=(8, 2))

    h[0] = -0.25 * (1.0 - xi[0]) * (1.0 - xi[1]) * (1.0 + xi[0] + xi[1])
    h[1] = 0.5 * (1.0 - xi[0]) * (1.0 + xi[0]) * (1.0 - xi[1])
    h[2] = -0.25 * (1.0 + xi[0]) * (1.0 - xi[1]) * (1.0 - xi[0] + xi[1])
    h[3] = 0.5 * (1.0 + xi[0]) * (1.0 + xi[1]) * (1.0 - xi[1])
    h[4] = -0.25 * (1.0 + xi[0]) * (1.0 + xi[1]) * (1.0 - xi[0] - xi[1])
    h[5] = 0.5 * (1.0 - xi[0]) * (1.0 + xi[0]) * (1.0 + xi[1])
    h[6] = -0.25 * (1.0 - xi[0]) * (1.0 + xi[1]) * (1.0 + xi[0] - xi[1])
    h[7] = 0.5 * (1.0 - xi[0]) * (1.0 + xi[1]) * (1.0 - xi[1])

    dhdxi[0, 0] = -0.25 * (-1.0 + xi[1]) * (2.0 * xi[0] + xi[1])
    dhdxi[1, 0] = xi[0] * (-1.0 + xi[1])
    dhdxi[2, 0] = 0.25 * (-1.0 + xi[1]) * (-2.0 * xi[0] + xi[1])
    dhdxi[3, 0] = -0.5 * (1.0 + xi[1]) * (-1.0 + xi[1])
    dhdxi[4, 0] = 0.25 * (1.0 + xi[1]) * (2.0 * xi[0] + xi[1])
    dhdxi[5, 0] = -xi[0] * (1.0 + xi[1])
    dhdxi[6, 0] = -0.25 * (1.0 + xi[1]) * (-2.0 * xi[0] + xi[1])
    dhdxi[7, 0] = 0.5 * (1.0 + xi[1]) * (-1.0 + xi[1])

    dhdxi[0, 1] = -0.25 * (-1.0 + xi[0]) * (xi[0] + 2.0 * xi[1])
    dhdxi[1, 1] = 0.5 * (1.0 + xi[0]) * (-1.0 + xi[0])
    dhdxi[2, 1] = 0.25 * (1.0 + xi[0]) * (-xi[0] + 2.0 * xi[1])
    dhdxi[3, 1] = -xi[1] * (1.0 + xi[0])
    dhdxi[4, 1] = 0.25 * (1.0 + xi[0]) * (xi[0] + 2.0 * xi[1])
    dhdxi[5, 1] = -0.5 * (1.0 + xi[0]) * (-1.0 + xi[0])
    dhdxi[6, 1] = -0.25 * (-1.0 + xi[0]) * (-xi[0] + 2.0 * xi[1])
    dhdxi[7, 1] = xi[1] * (-1.0 + xi[0])

    return h, dhdxi


def get_shape_tetra4(xi: ndarray) -> Tuple[ndarray, ndarray]:
    """
    四节点四面体单元
    """

    # 3
    # * **
    # *   * *
    # *     *  *
    # *       *   2
    # *        **  *
    # x2    *     * *
    # |  x1         **
    # 0--x0 * * * * * 1

    if len(xi) != 3:
        raise NotImplementedError(error_style(f'coordinate {xi} must be dimension 3'))

    h = empty(4)
    dhdxi = empty(shape=(4, 3))

    h[0] = 1.0 - xi[0] - xi[1] - xi[2]
    h[1] = xi[0]
    h[2] = xi[1]
    h[3] = xi[2]

    dhdxi[0, 0] = -1.0
    dhdxi[1, 0] = 1.0
    dhdxi[2, 0] = 0.0
    dhdxi[3, 0] = 0.0

    dhdxi[0, 1] = -1.0
    dhdxi[1, 1] = 0.0
    dhdxi[2, 1] = 1.0
    dhdxi[3, 1] = 0.0

    dhdxi[0, 2] = -1.0
    dhdxi[1, 2] = 0.0
    dhdxi[2, 2] = 0.0
    dhdxi[3, 2] = 1.0

    return h, dhdxi


def get_shape_hex8(xi: ndarray) -> Tuple[ndarray, ndarray]:
    """
    八节点六面体单元
    """

    #     7---------------6
    #    /|              /|
    #   / |     x2 x1   / |
    #  /  |     | /    /  |
    # 4---+-----|/----5   |
    # |   |     o--x0 |   |
    # |   3-----------+---2
    # |  /            |  /
    # | /             | /
    # |/              |/
    # 0---------------1

    if len(xi) != 3:
        raise NotImplementedError(error_style(f'coordinate {xi} must be dimension 3'))

    h = empty(8)
    dhdxi = empty(shape=(8, 3))

    h[0] = 0.125 * (1.0 - xi[0]) * (1.0 - xi[1]) * (1.0 - xi[2])
    h[1] = 0.125 * (1.0 + xi[0]) * (1.0 - xi[1]) * (1.0 - xi[2])
    h[2] = 0.125 * (1.0 + xi[0]) * (1.0 + xi[1]) * (1.0 - xi[2])
    h[3] = 0.125 * (1.0 - xi[0]) * (1.0 + xi[1]) * (1.0 - xi[2])
    h[4] = 0.125 * (1.0 - xi[0]) * (1.0 - xi[1]) * (1.0 + xi[2])
    h[5] = 0.125 * (1.0 + xi[0]) * (1.0 - xi[1]) * (1.0 + xi[2])
    h[6] = 0.125 * (1.0 + xi[0]) * (1.0 + xi[1]) * (1.0 + xi[2])
    h[7] = 0.125 * (1.0 - xi[0]) * (1.0 + xi[1]) * (1.0 + xi[2])

    dhdxi[0, 0] = -0.125 * (1.0 - xi[1]) * (1.0 - xi[2])
    dhdxi[1, 0] = 0.125 * (1.0 - xi[1]) * (1.0 - xi[2])
    dhdxi[2, 0] = 0.125 * (1.0 + xi[1]) * (1.0 - xi[2])
    dhdxi[3, 0] = -0.125 * (1.0 + xi[1]) * (1.0 - xi[2])
    dhdxi[4, 0] = -0.125 * (1.0 - xi[1]) * (1.0 + xi[2])
    dhdxi[5, 0] = 0.125 * (1.0 - xi[1]) * (1.0 + xi[2])
    dhdxi[6, 0] = 0.125 * (1.0 + xi[1]) * (1.0 + xi[2])
    dhdxi[7, 0] = -0.125 * (1.0 + xi[1]) * (1.0 + xi[2])

    dhdxi[0, 1] = -0.125 * (1.0 - xi[0]) * (1.0 - xi[2])
    dhdxi[1, 1] = -0.125 * (1.0 + xi[0]) * (1.0 - xi[2])
    dhdxi[2, 1] = 0.125 * (1.0 + xi[0]) * (1.0 - xi[2])
    dhdxi[3, 1] = 0.125 * (1.0 - xi[0]) * (1.0 - xi[2])
    dhdxi[4, 1] = -0.125 * (1.0 - xi[0]) * (1.0 + xi[2])
    dhdxi[5, 1] = -0.125 * (1.0 + xi[0]) * (1.0 + xi[2])
    dhdxi[6, 1] = 0.125 * (1.0 + xi[0]) * (1.0 + xi[2])
    dhdxi[7, 1] = 0.125 * (1.0 - xi[0]) * (1.0 + xi[2])

    dhdxi[0, 2] = -0.125 * (1.0 - xi[0]) * (1.0 - xi[1])
    dhdxi[1, 2] = -0.125 * (1.0 + xi[0]) * (1.0 - xi[1])
    dhdxi[2, 2] = -0.125 * (1.0 + xi[0]) * (1.0 + xi[1])
    dhdxi[3, 2] = -0.125 * (1.0 - xi[0]) * (1.0 + xi[1])
    dhdxi[4, 2] = 0.125 * (1.0 - xi[0]) * (1.0 - xi[1])
    dhdxi[5, 2] = 0.125 * (1.0 + xi[0]) * (1.0 - xi[1])
    dhdxi[6, 2] = 0.125 * (1.0 + xi[0]) * (1.0 + xi[1])
    dhdxi[7, 2] = 0.125 * (1.0 - xi[0]) * (1.0 + xi[1])

    return h, dhdxi


def get_gauss_points(dimension: int, order: int) -> Tuple[ndarray, ndarray]:
    xi, weight = leggauss(order)
    if dimension == 1:
        xi = xi.reshape(len(xi), -1)
        weight = weight.reshape(len(weight), -1)

    elif dimension == 2:
        xi1, xi2 = meshgrid(xi, xi)
        xi1 = xi1.ravel()
        xi2 = xi2.ravel()
        xi = column_stack((xi1, xi2))
        weight = outer(weight, weight)
        weight = weight.ravel()

    elif dimension == 3:
        xi1, xi2, xi3 = meshgrid(xi, xi, xi)
        xi1 = xi1.ravel()
        xi2 = xi2.ravel()
        xi3 = xi3.ravel()
        xi = column_stack((xi1, xi2, xi3))
        weight = outer(outer(weight, weight), weight)
        weight = weight.ravel()

    return xi, weight


def get_gauss_points_triangle(order: int) -> Tuple[ndarray, ndarray]:
    if order == 1:
        xi = [[1.0 / 3.0, 1.0 / 3.0]]
        weight = [0.5]

    elif order == 3:
        r1 = 1.0 / 6.0
        r2 = 2.0 / 3.0
        xi = [[r1, r1], [r2, r1], [r1, r2]]
        w1 = 1.0 / 6.0
        weight = [w1, w1, w1]

    elif order == 7:
        r1 = 0.5 * 0.1012865073235
        r2 = 0.5 * 0.7974269853531
        r4 = 0.5 * 0.4701420641051
        r6 = 0.0597158717898
        r7 = 1.0 / 3.0
        xi = [[r1, r1], [r2, r1], [r1, r2], [r4, r6], [r4, r4], [r6, r4], [r7, r7]]
        w1 = 0.1259391805448
        w4 = 0.1323941527885
        w7 = 0.225
        weight = [w1, w1, w1, w4, w4, w4, w7]

    else:
        raise NotImplementedError(error_style('Order must be 1, 3 or 7'))

    return array(xi), array(weight)


def get_gauss_points_tetra(order: int) -> Tuple[ndarray, ndarray]:
    if order == 1:
        third = 1.0 / 3.0
        xi = [[third, third, third]]
        weight = [0.5 * third]
    else:
        raise NotImplementedError(error_style('Only order 1 integration implemented'))

    return array(xi), array(weight)


def get_gauss_points_pyramid(order: int) -> Tuple[ndarray, ndarray]:
    if order == 1:
        xi = [[0., 0., -0.5]]
        weight = [128.0 / 27.0]
    else:
        raise NotImplementedError(error_style('Only order 1 integration implemented'))

    return array(xi), array(weight)


def get_default_element_type(node_coords: ndarray) -> str:
    num_element_nodes = node_coords.shape[0]
    dimension = node_coords.shape[1]

    if dimension == 1:
        if num_element_nodes == 2:
            return "line2"
        elif num_element_nodes == 3:
            return "line3"
        else:
            error_msg = f'No 1D element with {num_element_nodes} nodes available'
            raise NotImplementedError(error_style(error_msg))
    elif dimension == 2:
        if num_element_nodes == 3:
            return "tria3"
        elif num_element_nodes == 4:
            return "quad4"
        elif num_element_nodes == 6:
            return "tria6"
        elif num_element_nodes == 8:
            return "quad8"
        elif num_element_nodes == 9:
            return "quad9"
        else:
            error_msg = f'No 2D element with {num_element_nodes} nodes available'
            raise NotImplementedError(error_style(error_msg))
    elif dimension == 3:
        if num_element_nodes == 4:
            return "tetra4"
        elif num_element_nodes == 5:
            return "pyramid5"
        elif num_element_nodes == 6:
            return "prism6"
        elif num_element_nodes == 8:
            return "hex8"
        elif num_element_nodes == 18:
            return "prism18"
        elif num_element_nodes == 20:
            return "hex20"
        else:
            error_msg = f'No 3D element with {num_element_nodes} nodes available'
            raise NotImplementedError(error_style(error_msg))
    else:
        raise NotImplementedError(error_style(f'Unsupported dimension {dimension}'))


if __name__ == "__main__":
    # iso_element_shape = IsoElementShape('tria3')
    # iso_element_shape = IsoElementShape('quad4')
    # iso_element_shape = IsoElementShape('hex8')
    # iso_element_shape = IsoElementShape('quad8')
    # iso_element_shape = IsoElementShape('tetra4')
    # iso_element_shape = IsoElementShape('line2')
    # iso_element_shape = IsoElementShape('line3')
    iso_element_shape = IsoElementShape('empty')
    print(iso_element_shape.to_string())



from typing import Optional

from pyfem.io.Material import Material
from pyfem.materials.BaseMaterial import BaseMaterial
from pyfem.materials.ElasticIsotropic import ElasticIsotropic
from pyfem.utils.colors import error_style

material_data_dict = {
    'ElasticIsotropic': ElasticIsotropic
}


def get_material_data(material: Material, dimension: int, option: Optional[str] = None) -> BaseMaterial:
    class_name = f'{material.category}{material.type}'.strip().replace(' ', '')

    if class_name in material_data_dict:
        return material_data_dict[class_name](material, dimension, option)
    else:
        error_msg = f'{class_name} material is not supported.\n'
        error_msg += f'The allowed material types are {list(material_data_dict.keys())}.'
        raise NotImplementedError(error_style(error_msg))


if __name__ == "__main__":
    from pyfem.io.Properties import Properties

    props = Properties()
    props.read_file(r'F:\Github\pyfem\examples\rectangle\rectangle.toml')

    material_data = get_material_data(props.materials[0], 3)

    print(material_data.to_string())



import inspect
import time

from pyfem.utils.colors import BOLD, MAGENTA, YELLOW, BLUE, END


def show_running_time(func):
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        running_time = end_time - start_time
        print(BOLD + MAGENTA + f'{func} running time ' + END + YELLOW + f'= {running_time} s.' + END)
        return result

    return wrapper


def trace_calls(func):
    def wrapper(*args, **kwargs):
        call_stack = inspect.stack()
        call_frames = []
        for frame in call_stack[1:]:
            filename = frame.filename.split('src')[-1]
            call_frames.append((filename, frame.function, frame.lineno))
        print(f'{BLUE}{func} called from:{END}')
        for frame in call_frames:
            print(f'{YELLOW}\t{frame[0]}:{END}{frame[1]}: line {frame[2]}')
        return func(*args, **kwargs)

    return wrapper



from typing import Optional, List

from pyfem.utils.visualization import object_dict_to_string


class Dof:
    def __init__(self) -> None:
        self.names: List[str] = None  # type: ignore
        self.family: Optional[str] = None
        self.order: Optional[int] = None

    def to_string(self, level: int = 1) -> str:
        return object_dict_to_string(self, level)

    def show(self) -> None:
        print(self.to_string())


if __name__ == "__main__":
    dof = Dof()
    print(dof.__dict__.keys())
    print(dof)
    print(dof.to_string())



from typing import List, Dict

from numpy import repeat, array, ndarray, empty, zeros
from scipy.sparse import coo_matrix, csc_matrix  # type: ignore

from pyfem.bc.BaseBC import BaseBC
from pyfem.bc.get_bc_data import get_bc_data
from pyfem.elements.BaseElement import BaseElement
from pyfem.elements.IsoElementShape import IsoElementShape
from pyfem.elements.get_element_data import get_element_data
from pyfem.elements.get_iso_element_type import get_iso_element_type
from pyfem.io.Properties import Properties
from pyfem.materials.get_material_data import get_material_data
from pyfem.utils.visualization import object_dict_to_string_assembly
from pyfem.utils.wrappers import show_running_time

iso_element_shape_dict = {
    'line2': IsoElementShape('line2'),
    'line3': IsoElementShape('line3'),
    'tria3': IsoElementShape('tria3'),
    'quad4': IsoElementShape('quad4'),
    'quad8': IsoElementShape('quad8'),
    'tetra4': IsoElementShape('tetra4'),
    'hex8': IsoElementShape('hex8')
}


class Assembly:
    def __init__(self, props: Properties) -> None:
        self.total_dof_number: int = -1
        self.props: Properties = props
        self.materials_dict: Dict = {}
        self.sections_dict: Dict = {}
        self.section_of_element_set: Dict = {}
        self.element_data_list: List[BaseElement] = []
        self.bc_data_list: List[BaseBC] = []
        self.global_stiffness: csc_matrix = csc_matrix(0)
        self.fext: ndarray = empty(0)
        self.fint: ndarray = empty(0)
        self.dof_solution: ndarray = empty(0)
        self.field_variables: Dict[str, ndarray] = {}
        self.init_data_list()
        self.create_global_stiffness()
        self.apply_bcs()

    def to_string(self, level: int = 1) -> str:
        return object_dict_to_string_assembly(self, level)

    def show(self) -> None:
        print(self.to_string())

    @show_running_time
    def init_data_list(self) -> None:
        # 初始化 self.element_data_list
        elements = self.props.elements
        nodes = self.props.nodes
        sections = self.props.sections
        materials = self.props.materials
        dof = self.props.dof
        dimension = nodes.dimension
        self.total_dof_number = len(nodes) * len(dof.names)

        for material in materials:
            self.materials_dict[material.name] = material

        for section in sections:
            self.sections_dict[section.name] = section

        for element_set in elements.element_sets:
            for section in sections:
                if element_set in section.element_sets:
                    self.section_of_element_set[element_set] = section

        for element_set_name, element_ids in elements.element_sets.items():
            section = self.section_of_element_set[element_set_name]
            material = self.materials_dict[section.material_name]
            material_data = get_material_data(material=material,
                                              dimension=dimension,
                                              option=section.type)

            for element_id in element_ids:
                connectivity = elements[element_id]
                node_coords = array(nodes.get_items_by_ids(list(connectivity)))
                iso_element_type = get_iso_element_type(node_coords)
                iso_element_shape = iso_element_shape_dict[iso_element_type]
                element_data = get_element_data(element_id=element_id,
                                                iso_element_shape=iso_element_shape,
                                                connectivity=connectivity,
                                                node_coords=node_coords,
                                                section=section,
                                                dof=dof,
                                                material=material,
                                                material_data=material_data)

                element_data.assembly_conn = array(nodes.get_indices_by_ids(list(connectivity)))
                element_data.create_element_dof_ids()

                self.element_data_list.append(element_data)

        # 初始化 self.bc_data_list
        bcs = self.props.bcs
        for bc in bcs:
            bc_data = get_bc_data(bc=bc, dof=dof, nodes=nodes)
            self.bc_data_list.append(bc_data)

        # 初始化 RHS: self.fext
        self.fext = zeros(self.total_dof_number)

    def create_global_stiffness(self) -> None:
        val = []
        row = []
        col = []

        for element_data in self.element_data_list:
            element_dof_ids = element_data.element_dof_ids
            element_dof_number = element_data.element_dof_number
            row += [r for r in repeat(element_dof_ids, element_dof_number)]
            for _ in range(element_dof_number):
                col += [c for c in element_dof_ids]
            val += [v for v in element_data.stiffness.reshape(element_dof_number * element_dof_number)]

        self.global_stiffness = coo_matrix((array(val), (array(row), array(col))),
                                           shape=(self.total_dof_number, self.total_dof_number)).tocsc()

        # 以下代码采用 numpy.append 方法，处理可变对象时效率非常低，不建议使用

        # val = array([], dtype=float)
        # row = array([], dtype=int)
        # col = array([], dtype=int)
        #
        # for element_data in self.element_data_list:
        #     element_dof_ids = element_data.element_dof_ids
        #     element_dof_number = element_data.element_dof_number
        #     row = append(row, repeat(element_dof_ids, element_dof_number))
        #     for i in range(element_dof_number):
        #         col = append(col, element_dof_ids)
        #     val = append(val, element_data.stiffness.reshape(element_dof_number * element_dof_number))
        #
        # self.global_stiffness = coo_matrix((val, (row, col)), shape=(self.total_dof_number, self.total_dof_number))

    def apply_bcs(self) -> None:
        penalty = 1.0e16
        for bc_data in self.bc_data_list:
            for dof_id, dof_value in zip(bc_data.dof_ids, bc_data.dof_values):
                self.global_stiffness[dof_id, dof_id] += penalty
                self.fext[dof_id] += dof_value * penalty

    @show_running_time
    def update_field_variables(self, solution: ndarray) -> None:
        self.dof_solution = solution
        nodes_number = len(self.props.nodes)

        for element_data in self.element_data_list:
            element_data.update_field_variables(solution)

        for output in self.props.outputs:
            if output.type == 'vtk':
                for field_name in output.field_outputs:
                    self.field_variables[field_name] = zeros(nodes_number)
                    nodes_count = zeros(nodes_number)
                    for element_data in self.element_data_list:
                        assembly_conn = element_data.assembly_conn
                        self.field_variables[field_name][assembly_conn] += element_data.average_field_variables[
                            field_name]
                        nodes_count[assembly_conn] += 1.0
                    self.field_variables[field_name] = self.field_variables[field_name] / nodes_count

        # for field_name in self.props.outputs[0].field_outputs:
        #     self.field_variables[field_name] = zeros((nodes_number, 3))
        #     nodes_count = zeros(nodes_number)
        #     for element_data in self.element_data_list:
        #         assembly_conn = element_data.assembly_conn
        #         self.field_variables[field_name][assembly_conn] += element_data.average_field_variables[field_name]
        #         nodes_count[assembly_conn] += 1.0
        #     self.field_variables[field_name] = self.field_variables[field_name]/nodes_count.reshape(-1, 1)


@show_running_time
def main():
    from pyfem.solvers.get_solver_data import get_solver_data

    props = Properties()
    props.read_file(r'F:\Github\pyfem\examples\rectangle\rectangle.toml')
    props.verify()
    # props.show()
    assembly = Assembly(props)

    # x, info = gmres(A.toarray(), b, tol=1e-16)

    solver_data = get_solver_data(assembly, props.solver)

    solver_data.solve()

    solution = solver_data.solution

    assembly.update_field_variables(solution)

    import pprint
    pprint.pprint(assembly.field_variables)

    # print(props.nodes.node_sets['top'])


if __name__ == "__main__":
    main()



import os
from pathlib import Path
from typing import List, Any, Dict, Union

import meshio  # type: ignore
import numpy as np

from pyfem.utils.IntKeyDict import IntKeyDict
from pyfem.utils.logger import get_logger
from pyfem.utils.wrappers import show_running_time

logger = get_logger()


class NodeSet(IntKeyDict):

    def __init__(self) -> None:
        super().__init__()
        self.dimension: int = -1
        self.node_sets: Dict[str, List[int]] = {}

    def show(self) -> None:
        print(self.to_string(0))

    def to_string(self, level=1) -> str:
        msg = 'Number of nodes ............ %6d\n' % len(self)
        space = '   ' * level
        if len(self.node_sets) > 0:
            msg += space + '  Number of node_sets ........ %6d\n' % len(self.node_sets)
            msg += space + '  -----------------------------------\n'
            msg += space + '    name                       #nodes\n'
            msg += space + '    ---------------------------------\n'

            for name in self.node_sets:
                msg += space + '    %-16s           %6d\n' % (name, len(self.node_sets[name]))

        return msg[:-1]

    @show_running_time
    def read_gmsh_file(self, filename: Union[Path, str]) -> None:
        logger.info(f"Reading nodes from {filename}")

        mesh = meshio.read(filename, file_format='gmsh')

        # 可以通过以下命令查看支持的单元类型
        # print(meshio.gmsh.meshio_to_gmsh_type)
        # print(meshio.gmsh.gmsh_to_meshio_type)

        # keywords_1d = ['line']
        # keywords_2d = ['triangle', 'quad']
        keywords_3d = ['pyramid', 'hexahedron', 'wedge', 'tetra']

        self.dimension = 2

        # If any 3D mesh type is found, set the dimension to 3
        for cell_name, cell_dict in mesh.cell_sets_dict.items():
            for mesh_type in cell_dict.keys():
                if mesh_type in keywords_3d:
                    self.dimension = 3

        for node_id, coords in enumerate(mesh.points):
            self.add_item_by_id(node_id, coords[:self.dimension])

        for cell_name, cell_dict in mesh.cell_sets_dict.items():
            if cell_name != 'gmsh:bounding_entities':
                for mesh_type, element_ids in cell_dict.items():
                    for element_id in element_ids:
                        cell_nodes = mesh.cells_dict[mesh_type][element_id]
                        for node_id in cell_nodes:
                            self.add_to_sets_by_id(cell_name, node_id)

        for key in self.node_sets:
            self.node_sets[key] = list(set(self.node_sets[key]))

    def add_to_sets_by_id(self, node_set_name: str, node_id: int) -> None:
        if node_set_name not in self.node_sets:
            self.node_sets[node_set_name] = [int(node_id)]
        else:
            self.node_sets[node_set_name].append(int(node_id))

    def get_coords_by_ids(self, node_ids: List[int]) -> np.ndarray[Any, np.dtype[np.float64]]:
        return np.array(self.get_items_by_ids(node_ids))


if __name__ == "__main__":
    from pyfem.utils.logger import set_logger

    set_logger()

    nodes = NodeSet()
    os.chdir(r'/examples/rectangle')
    # nodes.read_gmsh_file('rectangle100.msh')
    nodes.read_gmsh_file('rectangle10000.msh')
    # print(nodes.dimension)
    # print(nodes.node_sets)
    print(nodes.get_coords_by_ids([0, 1, 2]))
    nodes.show()



from typing import List, Dict

from numpy import (dot, empty, array, ndarray)
from numpy.linalg import (det, inv)

from pyfem.elements.IsoElementShape import IsoElementShape
from pyfem.io.Dof import Dof
from pyfem.io.Material import Material
from pyfem.io.Section import Section
from pyfem.materials.BaseMaterial import BaseMaterial
from pyfem.utils.visualization import object_dict_to_string_ndarray
from pyfem.utils.wrappers import show_running_time


class BaseElement:
    def __init__(self, element_id: int, iso_element_shape: IsoElementShape, connectivity: ndarray,
                 node_coords: ndarray) -> None:
        self.element_id: int = element_id  # 用户自定义的节点编号
        self.iso_element_shape: IsoElementShape = iso_element_shape
        self.connectivity: ndarray = connectivity  # 对应用户定义的节点编号
        self.assembly_conn: ndarray = empty(0)  # 对应系统组装时的节点序号
        self.node_coords: ndarray = node_coords
        self.gp_jacobis: ndarray = empty(0)
        self.gp_jacobi_invs: ndarray = empty(0)
        self.gp_jacobi_dets: ndarray = empty(0)
        self.dof: Dof = None  # type: ignore
        self.dof_names: List[str] = []
        self.element_dof_ids: List[int] = []  # 对应系统组装时的自由度序号
        self.element_dof_values: ndarray = empty(0)  # 对应系统组装时的自由度的值
        self.element_dof_number: int = 0  # 单元自由度总数
        self.material: Material = None  # type: ignore
        self.section: Section = None  # type: ignore
        self.material_data: BaseMaterial = None  # type: ignore
        self.stiffness: ndarray = empty(0)
        self.gp_state_variables: ndarray = empty(0)
        self.field_variable_dict: Dict[str, List[str]] = {}
        self.gp_field_variables: Dict[str, ndarray] = {}
        self.average_field_variables: Dict[str, ndarray] = {}
        self.cal_jacobi()

    def to_string(self, level: int = 1) -> str:
        return object_dict_to_string_ndarray(self, level)

    def show(self) -> None:
        print(self.to_string())

    def cal_jacobi(self) -> None:
        """
        通过矩阵乘法计算每个积分点上的Jacobi矩阵。
        """

        # 以下代码为采用for循环的计算方法，结构清晰，但计算效率较低
        # self.jacobi = []
        # self.jacobi_inv = []
        # self.jacobi_det = []
        # for gp_shape_gradient in self.iso_element_shape.gp_shape_gradients:
        #     jacobi = dot(self.node_coords.transpose(), gp_shape_gradient)
        #     self.jacobi.append(jacobi)
        #     self.jacobi_inv.append(inv(jacobi))
        #     self.jacobi_det.append(det(jacobi))
        # self.jacobi = array(self.jacobi)
        # self.jacobi_inv = array(self.jacobi_inv)
        # self.jacobi_det = array(self.jacobi_det)

        # 以下代码为采用numpy高维矩阵乘法的计算方法，计算效率高，但要注意矩阵维度的变化
        self.gp_jacobis = dot(self.node_coords.transpose(), self.iso_element_shape.gp_shape_gradients).swapaxes(0, 1)
        self.gp_jacobi_invs = inv(self.gp_jacobis)
        self.gp_jacobi_dets = det(self.gp_jacobis)

    def create_element_dof_ids(self) -> None:
        for node_index in self.assembly_conn:
            for dof_id, _ in enumerate(self.dof_names):
                self.element_dof_ids.append(node_index * len(self.dof_names) + dof_id)

    def update_field_variables(self, solution: ndarray) -> None:
        pass


@show_running_time
def main():
    from pyfem.assembly.Assembly import iso_element_shape_dict
    from pyfem.io.Properties import Properties

    props = Properties()
    props.read_file(r'F:\Github\pyfem\examples\rectangle\rectangle.toml')

    elements = props.elements
    nodes = props.nodes

    base_elements = []

    for element_id, connectivity in elements.items():
        node_coords = array(nodes.get_items_by_ids(list(connectivity)))
        base_element = BaseElement(element_id, iso_element_shape_dict['quad4'], connectivity, node_coords)
        base_elements.append(base_element)

    print(base_elements[0].to_string())


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""

"""
from pathlib import Path
from typing import Union

from pyfem.assembly.Assembly import Assembly
from pyfem.io.Properties import Properties
from pyfem.io.write_vtk import write_vtk
from pyfem.solvers.get_solver_data import get_solver_data


class Job:
    def __init__(self, filename: Union[Path, str]) -> None:
        input_file = Path(filename)
        if input_file.is_absolute():
            abs_input_file = input_file
        else:
            abs_input_file = Path.cwd().joinpath(input_file)
        self.props = Properties()
        self.props.read_file(abs_input_file)
        self.assembly = Assembly(self.props)
        self.solver_data = get_solver_data(self.assembly, self.props.solver)

    def run(self):
        self.solver_data.run()
        write_vtk(self.props, self.assembly)



from numpy import empty
from scipy.sparse.linalg import spsolve  # type: ignore

from pyfem.assembly.Assembly import Assembly
from pyfem.io.Solver import Solver
from pyfem.solvers.BaseSolver import BaseSolver
from pyfem.utils.wrappers import show_running_time


class LinearSolver(BaseSolver):
    def __init__(self, assembly: Assembly, solver: Solver) -> None:
        super().__init__(assembly, solver)
        self.assembly: Assembly = assembly
        self.solver: Solver = solver
        self.solution = empty(0)

    def run(self) -> None:
        self.solve()
        self.update_field_variables()

    @show_running_time
    def solve(self) -> None:
        A = self.assembly.global_stiffness
        rhs = self.assembly.fext
        x = spsolve(A, rhs)
        self.solution = x

    def update_field_variables(self) -> None:
        self.assembly.update_field_variables(self.solution)


if __name__ == "__main__":
    pass



from numpy import (empty, ndarray)

from pyfem.io.BC import BC
from pyfem.io.Dof import Dof
from pyfem.mesh.NodeSet import NodeSet
from pyfem.utils.visualization import object_dict_to_string_ndarray


class BaseBC:
    def __init__(self, bc: BC, dof: Dof, nodes: NodeSet) -> None:
        self.bc: BC = bc
        self.dof: Dof = dof
        self.nodes: NodeSet = nodes
        self.bc_node_ids: ndarray = empty(0)
        self.bc_element_ids: ndarray = empty(0)
        self.dof_ids: ndarray = empty(0)
        self.dof_values: ndarray = empty(0)

    def to_string(self, level: int = 1) -> str:
        return object_dict_to_string_ndarray(self, level)

    def show(self) -> None:
        print(self.to_string())


if __name__ == "__main__":
    pass



from pathlib import Path
from typing import Dict, List, Any, Tuple, Union

try:
    import tomllib  # type: ignore
except ModuleNotFoundError:
    import tomli as tomllib

from pyfem.io.Dof import Dof
from pyfem.io.Section import Section
from pyfem.io.Material import Material
from pyfem.io.Mesh import Mesh
from pyfem.io.BC import BC
from pyfem.io.Solver import Solver
from pyfem.io.Output import Output
from pyfem.mesh.NodeSet import NodeSet
from pyfem.mesh.ElementSet import ElementSet
from pyfem.utils.colors import CYAN, MAGENTA, BLUE, END, error_style, info_style


class Properties:
    """
    Properties类用于解析配置文件中定义的属性。
    当 self.is_read_only = True 时：

    1. Properties 类的所有属性在首次被赋非None值后不能再被修改和删除，

    2. 此时许可的属性关键字存储在self.slots中。
    """
    is_read_only: bool = True
    slots: Tuple = (
        'work_path', 'input_file', 'toml', 'title', 'mesh', 'dof', 'materials', 'sections', 'bcs',
        'solver',
        'outputs', 'nodes', 'elements')

    def __init__(self) -> None:
        self.work_path: Path = None  # type: ignore
        self.input_file: Path = None  # type: ignore
        self.toml: Dict = None  # type: ignore
        self.title: str = None  # type: ignore
        self.mesh: Mesh = None  # type: ignore
        self.dof: Dof = None  # type: ignore
        self.materials: List[Material] = None  # type: ignore
        self.sections: List[Section] = None  # type: ignore
        self.bcs: List[BC] = None  # type: ignore
        self.solver: Solver = None  # type: ignore
        self.outputs: List[Output] = None  # type: ignore
        self.nodes: NodeSet = None  # type: ignore
        self.elements: ElementSet = None  # type: ignore

    def __setattr__(self, key, value):
        if self.is_read_only:
            if key not in self.slots:
                error_msg = f'{key} is not an allowable attribute keyword of {type(self).__name__}'
                raise AttributeError(error_style(error_msg))
            elif hasattr(self, key) and self.__getattribute__(key) is not None:
                error_msg = f'attribute {type(self).__name__}.{key} is READ ONLY'
                raise PermissionError(error_style(error_msg))
            else:
                super().__setattr__(key, value)
        else:
            super().__setattr__(key, value)

    def __delattr__(self, key):
        if self.is_read_only:
            error_msg = f'attribute {type(self).__name__}.{key} is READ ONLY'
            raise PermissionError(error_style(error_msg))
        else:
            super().__delattr__(key)

    def verify(self) -> None:
        print(info_style('Verifying the input ...'))
        is_error = False
        error_msg = '\nInput error:\n'
        for key in self.slots[2:]:  # 忽略这2个关键字：'work_path', 'input_file'，它们不是在.toml中定义的
            if self.__getattribute__(key) is None:
                is_error = True
                error_msg += f'  - {key} is missing\n'

        if is_error:
            error_msg += f'Please check the input file {self.input_file}'
            raise NotImplementedError(error_style(error_msg))

    def show(self) -> None:
        for key, item in self.__dict__.items():
            print()
            print(CYAN + f'+-{key}' + END)
            print(MAGENTA + f'  |- {type(item)}' + END)
            if hasattr(item, 'to_string'):
                print(f'  |- {item.to_string()}')
            else:
                print(f'  |- {item}')
            if isinstance(item, list):
                for i, it in enumerate(item):
                    print(BLUE + f'    |-{i}-{it.to_string(level=3)}' + END)

    def set_toml(self, toml: Dict) -> None:
        self.toml = toml

    def set_title(self, title: str) -> None:
        self.title = title

    def set_mesh(self, mesh_dict: Dict) -> None:
        self.mesh = Mesh()
        allowed_keys = self.mesh.__dict__.keys()
        for key, item in mesh_dict.items():
            if key in allowed_keys:
                self.mesh.__setattr__(key, item)
            else:
                raise AttributeError(self.key_error_message(key, self.mesh))
        if self.mesh.type == 'gmsh':
            self.set_nodes_from_gmsh()
            self.set_elements_from_gmsh()

    def set_dofs(self, dofs_dict: Dict) -> None:
        self.dof = Dof()
        allowed_keys = self.dof.__dict__.keys()
        for key, item in dofs_dict.items():
            if key in allowed_keys:
                self.dof.__setattr__(key, item)
            else:
                raise AttributeError(self.key_error_message(key, self.dof))

    def set_solver(self, solver_dict: Dict) -> None:
        self.solver = Solver()
        allowed_keys = self.solver.__dict__.keys()
        for key, item in solver_dict.items():
            if key in allowed_keys:
                self.solver.__setattr__(key, item)
            else:
                raise AttributeError(self.key_error_message(key, self.solver))

    def set_materials(self, materials_list: List) -> None:
        self.materials = []
        for material_dict in materials_list:
            material = Material()
            allowed_keys = material.__dict__.keys()
            for key, item in material_dict.items():
                if key in allowed_keys:
                    material.__setattr__(key, item)
                else:
                    raise AttributeError(self.key_error_message(key, material))
            self.materials.append(material)

    def set_sections(self, sections_list: List) -> None:
        self.sections = []
        for section_dict in sections_list:
            section = Section()
            allowed_keys = section.__dict__.keys()
            for key, item in section_dict.items():
                if key in allowed_keys:
                    section.__setattr__(key, item)
                else:
                    raise AttributeError(self.key_error_message(key, section))
            self.sections.append(section)

    def set_bcs(self, bcs_list: List) -> None:
        self.bcs = []
        for bc_dict in bcs_list:
            bc = BC()
            allowed_keys = bc.__dict__.keys()
            for key, item in bc_dict.items():
                if key in allowed_keys:
                    bc.__setattr__(key, item)
                else:
                    raise AttributeError(self.key_error_message(key, bc))
            self.bcs.append(bc)

    def set_outputs(self, outputs_list: List) -> None:
        self.outputs = []
        for output_dict in outputs_list:
            output = Output()
            allowed_keys = output.__dict__.keys()
            for key, item in output_dict.items():
                if key in allowed_keys:
                    output.__setattr__(key, item)
                else:
                    raise AttributeError(self.key_error_message(key, output))
            self.outputs.append(output)

    def set_nodes_from_gmsh(self):
        gmsh_path = Path(self.mesh.file)
        if gmsh_path.is_absolute():  # 判断 self.mesh.file 是不是绝对路径
            abs_gmsh_file = gmsh_path
        else:  # 如果 self.mesh.file 不是绝对路径，则用工作目录 self.work_path 补全为绝对路径
            abs_gmsh_file = self.work_path.joinpath(gmsh_path)
        self.nodes = NodeSet()
        self.nodes.read_gmsh_file(abs_gmsh_file)
        self.nodes.update_indices()

    def set_elements_from_gmsh(self):
        gmsh_path = Path(self.mesh.file)
        if gmsh_path.is_absolute():  # 判断 self.mesh.file 是不是绝对路径
            abs_gmsh_file = gmsh_path
        else:  # 如果 self.mesh.file 不是绝对路径，则用工作目录 self.work_path 补全为绝对路径
            abs_gmsh_file = self.work_path.joinpath(gmsh_path)
        self.elements = ElementSet()
        self.elements.read_gmsh_file(abs_gmsh_file, self.sections)

    @staticmethod
    def key_error_message(key: Any, obj: Any) -> str:
        return error_style(f'{key} is not an allowable attribute keyword of {type(obj).__name__}')

    def read_file(self, filename: Union[Path, str]) -> None:
        """
        读取 .toml 格式的配置文件。
        """
        self.input_file = Path(filename)
        self.work_path = self.input_file.parent

        with open(self.input_file, "rb") as f:
            toml = tomllib.load(f)
            self.set_toml(toml)

        toml_keys = self.toml.keys()
        allowed_keys = self.__dict__.keys()

        for key in toml_keys:
            if key not in allowed_keys:
                error_msg = f'{key} is not an allowable attribute keyword of {type(self).__name__}\n'
                error_msg += f'Please check the file {self.input_file}'
                raise AttributeError(error_style(error_msg))

        if 'title' in toml_keys:
            title = self.toml['title']
            self.set_title(title)

        if 'sections' in toml_keys:
            sections_list = self.toml['sections']
            self.set_sections(sections_list)

        if 'mesh' in toml_keys:
            mesh_dict = self.toml['mesh']
            self.set_mesh(mesh_dict)

        if 'dof' in toml_keys:
            dofs_dict = self.toml['dof']
            self.set_dofs(dofs_dict)

        if 'materials' in toml_keys:
            materials_list = self.toml['materials']
            self.set_materials(materials_list)

        if 'bcs' in toml_keys:
            bcs_list = self.toml['bcs']
            self.set_bcs(bcs_list)

        if 'solver' in toml_keys:
            solver_dict = self.toml['solver']
            self.set_solver(solver_dict)

        if 'outputs' in toml_keys:
            outputs_dict = self.toml['outputs']
            self.set_outputs(outputs_dict)

        self.verify()


if __name__ == "__main__":
    props = Properties()
    # props.show()
    props.read_file(r'F:\Github\pyfem\examples\rectangle\rectangle.toml')
    props.show()



from typing import Optional, List, Tuple, Dict

from pyfem.utils.colors import error_style
from pyfem.utils.visualization import object_dict_to_string


class Material:
    """
    Material类用于存储配置文件中定义的材料属性。
    当 self.is_read_only = True 时：

    1. Material 类的所有属性在首次被赋非None值后不能再被修改和删除，

    2. 此时许可的属性关键字存储在self.slots中。
    """
    is_read_only: bool = True
    slots: Tuple = ('name', 'category', 'type', 'data')
    allowed_keys_values: Dict = {
        'category': [None, 'Elastic'],
        'type': [None, 'Isotropic']
    }

    def __init__(self) -> None:
        self.name: Optional[str] = None
        self.category: Optional[str] = None
        self.type: Optional[str] = None
        self.data: Optional[List[float]] = None

    def __setattr__(self, key, value) -> None:
        if self.is_read_only:
            if key not in self.slots:
                error_msg = f'{key} is not an allowable attribute keyword of {type(self).__name__}'
                raise AttributeError(error_style(error_msg))

            elif hasattr(self, key) and self.__getattribute__(key) is not None:
                error_msg = f'attribute {type(self).__name__}.{key} is READ ONLY'
                raise PermissionError(error_style(error_msg))

            else:
                for allowed_key, allowed_values in self.allowed_keys_values.items():
                    if key == allowed_key and value not in allowed_values:
                        error_msg = f'{value} is unsupported for {type(self).__name__}.{key}\n'
                        error_msg += f'The allowed values of {type(self).__name__}.{key} are {self.allowed_keys_values[key]}'
                        raise AttributeError(error_style(error_msg))

                super().__setattr__(key, value)
        else:
            super().__setattr__(key, value)

    def to_string(self, level: int = 1) -> str:
        return object_dict_to_string(self, level)

    def show(self) -> None:
        print(self.to_string())


if __name__ == "__main__":
    material = Material()
    print(material.__dict__.keys())
    print(material)
    print(material.to_string())



from typing import Optional

from pyfem.utils.visualization import object_dict_to_string


class Mesh:
    def __init__(self) -> None:
        self.type: Optional[str] = None
        self.file: Optional[str] = None

    def to_string(self, level: int = 1) -> str:
        return object_dict_to_string(self, level)

    def show(self) -> None:
        print(self.to_string())


if __name__ == "__main__":
    mesh = Mesh()
    print(mesh.__dict__.keys())
    print(mesh)
    print(mesh.to_string())



import logging
from typing import Optional

import colorlog


def set_logger(log_file: Optional[str] = None, level: int = logging.INFO) -> Optional[logging.Logger]:
    """
    设置 logging 模块的基础配置，并返回一个 logger 对象。
    :param log_file: 日志文件名，None 表示将日志输出到控制台
    :param level: 输出日志的最低级别，默认是 INFO 级别
    """
    # 创建 logger 对象
    logger = logging.getLogger()

    # 设置日志级别
    logger.setLevel(level)

    # 创建格式化器
    if log_file:
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    else:
        formatter = colorlog.ColoredFormatter('%(log_color)s%(message)s')

    # 创建处理器
    if log_file:
        # 日志输出到文件
        file_handler = logging.FileHandler(log_file)
        file_handler.setLevel(level)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
    else:
        # 日志输出到控制台
        console_handler = logging.StreamHandler()
        console_handler.setLevel(level)
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)

    return logger


def get_logger():
    return logging.getLogger()


if __name__ == "__main__":
    set_logger()
    logger = get_logger()
    logger.info("Information message")
    logger.warning("Warning message")
    logger.error("Error message")



from numpy import array, empty, zeros, dot, ndarray, average

from pyfem.elements.BaseElement import BaseElement
from pyfem.elements.IsoElementShape import IsoElementShape
from pyfem.io.Dof import Dof
from pyfem.io.Material import Material
from pyfem.io.Section import Section
from pyfem.materials.BaseMaterial import BaseMaterial
from pyfem.utils.colors import error_style


class SolidPlaneSmallStrain(BaseElement):

    def __init__(self, element_id: int,
                 iso_element_shape: IsoElementShape,
                 connectivity: ndarray,
                 node_coords: ndarray,
                 section: Section,
                 dof: Dof,
                 material: Material,
                 material_data: BaseMaterial) -> None:

        super().__init__(element_id, iso_element_shape, connectivity, node_coords)
        self.dof = dof
        self.dof_names = ['u1', 'u2']
        self.field_variable_dict = {
            'strain': ['E11', 'E22', 'E12'],
            'stress': ['S11', 'S22', 'S12']
        }
        if dof.names != self.dof_names:
            error_msg = f'{dof.names} is not the supported dof of {type(self).__name__} element'
            raise NotImplementedError(error_style(error_msg))
        self.element_dof_number = len(self.dof_names) * self.iso_element_shape.nodes_number
        self.material = material
        self.section = section
        self.material_data = material_data
        self.gp_b_matrices = empty(0)
        self.stiffness = empty(0)
        self.update_gp_b_matrices()
        self.update_stiffness()

    def update_gp_b_matrices(self) -> None:

        self.gp_b_matrices = zeros(shape=(self.iso_element_shape.gp_number, 3, self.element_dof_number))

        for igp, (gp_shape_gradient, gp_jacobi_inv) in \
                enumerate(zip(self.iso_element_shape.gp_shape_gradients, self.gp_jacobi_invs)):
            gp_dhdx = dot(gp_shape_gradient, gp_jacobi_inv)
            for i, val in enumerate(gp_dhdx):
                self.gp_b_matrices[igp, 0, i * 2] = val[0]
                self.gp_b_matrices[igp, 1, i * 2 + 1] = val[1]
                self.gp_b_matrices[igp, 2, i * 2] = val[1]
                self.gp_b_matrices[igp, 2, i * 2 + 1] = val[0]

    def update_stiffness(self) -> None:
        self.stiffness = zeros(shape=(self.element_dof_number, self.element_dof_number))

        ddsdde = self.material_data.get_tangent()
        gp_weights = self.iso_element_shape.gp_weights
        gp_jacobi_dets = self.gp_jacobi_dets
        gp_b_matrices = self.gp_b_matrices
        gp_number = self.iso_element_shape.gp_number

        for i in range(gp_number):
            self.stiffness += dot(gp_b_matrices[i].transpose(), dot(ddsdde, gp_b_matrices[i])) * gp_weights[i] * \
                              gp_jacobi_dets[i]

    def update_field_variables(self, solution: ndarray) -> None:
        gp_b_matrices = self.gp_b_matrices
        gp_number = self.iso_element_shape.gp_number
        ddsdde = self.material_data.get_tangent()

        self.element_dof_values = solution[self.element_dof_ids]

        gp_strains = []
        gp_stresses = []
        for i in range(gp_number):
            gp_strain = dot(gp_b_matrices[i], self.element_dof_values)
            gp_stress = dot(ddsdde, gp_strain)
            gp_strains.append(gp_strain)
            gp_stresses.append(gp_stress)

        self.gp_field_variables['strain'] = array(gp_strains)
        self.gp_field_variables['stress'] = array(gp_stresses)

        # self.average_field_variables['strain'] = average(self.gp_field_variables['strain'], axis=0)
        # self.average_field_variables['stress'] = average(self.gp_field_variables['stress'], axis=0)

        average_strain = average(self.gp_field_variables['strain'], axis=0)
        average_stress = average(self.gp_field_variables['stress'], axis=0)

        self.average_field_variables['E11'] = average_strain[0]
        self.average_field_variables['E22'] = average_strain[1]
        self.average_field_variables['E12'] = average_strain[2]
        self.average_field_variables['S11'] = average_stress[0]
        self.average_field_variables['S22'] = average_stress[1]
        self.average_field_variables['S12'] = average_stress[2]


if __name__ == "__main__":
    pass



from typing import Optional, Tuple

from numpy import array, outer, diag, float64, ndarray

from pyfem.io.Material import Material
from pyfem.materials.BaseMaterial import BaseMaterial
from pyfem.utils.colors import error_style


class ElasticIsotropic(BaseMaterial):
    allowed_option = ['PlaneStress', 'PlaneStrain', None]

    def __init__(self, material: Material, dimension: int, option: Optional[str] = None) -> None:
        super().__init__(material, dimension, option)
        self.create_tangent()

    def create_tangent(self):
        young = self.material.data[0]
        poisson = self.material.data[1]

        if self.option in self.allowed_option:
            if self.dimension == 3:
                self.option = None
            self.ddsdde = get_stiffness_from_young_poisson(self.dimension, young, poisson, self.option)
        else:
            error_msg = f'{self.option} is not the allowed options {self.allowed_option}'
            raise NotImplementedError(error_style(error_msg))


def get_lame_from_young_poisson(young: float, poisson: float, plane: Optional[str]) -> Tuple[float, float]:
    r"""
    Compute Lamé parameters from Young's modulus and Poisson's ratio.

    The relationship between Lamé parameters and Young's modulus, Poisson's
    ratio (see [1],[2]):

    .. math::
        \lambda = {\nu E \over (1+\nu)(1-2\nu)},\qquad \mu = {E \over 2(1+\nu)}

    The plain stress hypothesis:

    .. math::
       \bar\lambda = {2\lambda\mu \over \lambda + 2\mu}

    [1] I.S. Sokolnikoff: Mathematical Theory of Elasticity. New York, 1956.

    [2] T.J.R. Hughes: The Finite Element Method, Linear Static and Dynamic
    Finite Element Analysis. New Jersey, 1987.
    """
    mu = young / (2.0 * (1.0 + poisson))
    lam = young * poisson / ((1.0 + poisson) * (1.0 - 2.0 * poisson))

    if plane == 'PlaneStress':
        lam = 2 * lam * mu / (lam + 2 * mu)

    return lam, mu


def get_stiffness_from_lame(dim: int, lam: float, mu: float) -> ndarray:
    r"""
    Compute stiffness tensor corresponding to Lamé parameters.

    .. math::
        {\bf D}_{(2D)} = \begin{bmatrix} \lambda + 2\mu & \lambda & 0\\
        \lambda & \lambda + 2\mu & 0\\ 0 & 0 & \mu \end{bmatrix}

    .. math::
        {\bf D}_{(3D)} = \begin{bmatrix} \lambda + 2\mu & \lambda &
        \lambda & 0 & 0 & 0\\ \lambda & \lambda + 2\mu & \lambda & 0 & 0 & 0 \\
        \lambda & \lambda & \lambda + 2\mu & 0 & 0 & 0 \\ 0 & 0 & 0 & \mu & 0 &
        0 \\ 0 & 0 & 0 & 0 & \mu & 0 \\ 0 & 0 & 0 & 0 & 0 & \mu\\ \end{bmatrix}
    """
    sym = (dim + 1) * dim // 2
    o = array([1.] * dim + [0.] * (sym - dim), dtype=float64)
    oot = outer(o, o)
    do1 = diag(o + 1.0)

    lam_array = array(lam)[..., None, None]
    mu_array = array(mu)[..., None, None]
    return lam_array * oot + mu_array * do1


def get_stiffness_from_young_poisson(dim: int, young: float, poisson: float, plane: Optional[str]) -> ndarray:
    """
    Compute stiffness tensor corresponding to Young's modulus and Poisson's
    ratio.
    """

    lam, mu = get_lame_from_young_poisson(young, poisson, plane)

    return get_stiffness_from_lame(dim, lam, mu)


def get_stiffness_from_lame_mixed(dim: int, lam: float, mu: float) -> ndarray:
    r"""
    Compute stiffness tensor corresponding to Lamé parameters for mixed
    formulation.

    .. math::
        {\bf D}_{(2D)} = \begin{bmatrix} \widetilde\lambda + 2\mu &
        \widetilde\lambda & 0\\ \widetilde\lambda & \widetilde\lambda + 2\mu &
        0\\ 0 & 0 & \mu \end{bmatrix}

    .. math::
        {\bf D}_{(3D)} = \begin{bmatrix} \widetilde\lambda + 2\mu &
        \widetilde\lambda & \widetilde\lambda & 0 & 0 & 0\\ \widetilde\lambda &
        \widetilde\lambda + 2\mu & \widetilde\lambda & 0 & 0 & 0 \\
        \widetilde\lambda & \widetilde\lambda & \widetilde\lambda + 2\mu & 0 &
        0 & 0 \\ 0 & 0 & 0 & \mu & 0 & 0 \\ 0 & 0 & 0 & 0 & \mu & 0 \\ 0 & 0 &
        0 & 0 & 0 & \mu\\ \end{bmatrix}

    where

    .. math::
       \widetilde\lambda = -{2\over 3} \mu
    """
    lam = - 2.0 / 3.0 * mu

    return get_stiffness_from_lame(dim, lam, mu)


def get_stiffness_from_young_poisson_mixed(dim: int, young: float, poisson: float, plane) -> ndarray:
    """
    Compute stiffness tensor corresponding to Young's modulus and Poisson's
    ratio for mixed formulation.
    """
    lam, mu = get_lame_from_young_poisson(young, poisson, plane)

    return get_stiffness_from_lame_mixed(dim, lam, mu)


def get_bulk_from_lame(lam: float, mu: float) -> float:
    r"""
    Compute bulk modulus from Lamé parameters.

    .. math::
        \gamma = \lambda + {2 \over 3} \mu
    """
    return lam + 2.0 * mu / 3.0


def get_bulk_from_young_poisson(young: float, poisson: float, plane: Optional[str]) -> float:
    """
    Compute bulk modulus corresponding to Young's modulus and Poisson's ratio.
    """
    lam, mu = get_lame_from_young_poisson(young, poisson, plane)

    return get_bulk_from_lame(lam, mu)


def get_lame_from_stiffness(stiffness: ndarray, plane: Optional[str]) -> Tuple[float, float]:
    """
    Compute Lamé parameters from an isotropic stiffness tensor.
    """
    lam = float(stiffness[..., 0, 1])
    mu = float(stiffness[..., -1, -1])
    if plane == 'PlaneStress':
        lam = - 2.0 * mu * lam / (lam - 2.0 * mu)

    return lam, mu


def get_young_poisson_from_stiffness(stiffness: ndarray, plane: Optional[str]) -> Tuple[float, float]:
    """
    Compute Young's modulus and Poisson's ratio from an isotropic stiffness
    tensor.
    """
    lam, mu = get_lame_from_stiffness(stiffness, plane)
    young = (3 * lam * mu + 2 * mu ** 2) / (lam + mu)
    poisson = lam / (2 * lam + 2 * mu)

    return young, poisson


if __name__ == "__main__":
    from pyfem.io.Properties import Properties

    props = Properties()
    props.read_file(r'F:\Github\pyfem\examples\rectangle\rectangle.toml')

    material_data = ElasticIsotropic(props.materials[0], 3)
    print(material_data.to_string())



import os
from copy import deepcopy
from pathlib import Path
from typing import List, Any, Dict, Union

import meshio  # type: ignore
import numpy as np

from pyfem.io.Section import Section
from pyfem.utils.IntKeyDict import IntKeyDict
from pyfem.utils.logger import get_logger
from pyfem.utils.wrappers import show_running_time

logger = get_logger()


class ElementSet(IntKeyDict):

    def __init__(self) -> None:
        super().__init__()
        self.element_sets: Dict[str, List[int]] = {}

    def show(self) -> None:
        print(self.to_string(0))

    def to_string(self, level: int = 1) -> str:
        msg = 'Number of elements ......... %6d\n' % len(self)
        space = '   ' * level
        if len(self.element_sets) > 0:
            msg += space + '  Number of element_sets ..... %6d\n' % len(self.element_sets)
            msg += space + '  -----------------------------------\n'
            msg += space + '    name                       #elems\n'
            msg += space + '    ---------------------------------\n'

            for name in self.element_sets:
                msg += space + '    %-16s           %6d\n' % (name, len(self.element_sets[name]))

        return msg[:-1]

    def get_dof_types(self) -> List[str]:
        dof_types = []
        for element in self:
            for dof_type in element.dof_types:
                if dof_type not in dof_types:
                    dof_types.append(dof_type)
        return dof_types

    @show_running_time
    def read_gmsh_file(self, filename: Union[Path, str], sections: List[Section]) -> None:
        mesh = meshio.read(filename, file_format="gmsh")

        assigned_element_set = []
        for section in sections:
            assigned_element_set += section.element_sets

        assembly_element_id = 0
        for cell_name, cell_dict in mesh.cell_sets_dict.items():
            if cell_name != 'gmsh:bounding_entities' and cell_name in assigned_element_set:
                for mesh_type, element_ids in cell_dict.items():
                    for element_id in element_ids:
                        connectivity = deepcopy(mesh.cells_dict[mesh_type][element_id])
                        self.add_item_by_element_id(assembly_element_id, cell_name, connectivity)
                        assembly_element_id += 1

    def add_item_by_element_id(self, element_id: int, element_set_name: str,
                               connectivity: np.ndarray[Any, np.dtype[np.int64]]) -> None:
        self.add_item_by_id(element_id, connectivity)
        self.add_to_sets_by_id(element_set_name, element_id)

    def add_to_sets_by_id(self, element_set_name: str, element_id: int) -> None:
        if element_set_name not in self.element_sets:
            self.element_sets[element_set_name] = [element_id]
        else:
            self.element_sets[element_set_name].append(element_id)


if __name__ == "__main__":
    from pyfem.io.Properties import Properties

    props = Properties()
    props.read_file(r'F:\Github\pyfem\examples\rectangle\rectangle.toml')

    os.chdir(r'/examples/rectangle')

    elements = ElementSet()
    elements.read_gmsh_file('rectangle100.msh', props.sections)
    elements.show()
